# SPDX-License-Identifier: LGPL-2.1-or-later

import abc


class XDRReader:
    def __init__(self, fp):
        self.fp = fp
        self.lookahead = ""
        self.lookbehind = ""
        self.line = 1
        self.column = 0

    def _read(self):
        if len(self.lookahead) > 0:
            c = self.lookahead[0:1]
            self.lookahead = self.lookahead[1:]
            return c
        return self.fp.read(1)

    def peek(self, skip=0):
        need = 1 + skip
        if len(self.lookahead) < need:
            self.lookahead = self.lookahead + self.fp.read(need - len(self.lookahead))
        if len(self.lookahead) < need:
            return None

        return self.lookahead[skip : skip + 1]

    def last(self, skip=0):
        if (skip + 1) > len(self.lookbehind):
            return None
        return self.lookbehind[skip]

    def next(self):
        c = self._read()
        line = self.line
        column = self.column
        if c == "\n":
            self.line = self.line + 1
            self.column = 0
        else:
            self.column = self.column + 1
        self.lookbehind = c + self.lookbehind
        if len(self.lookbehind) > 2:
            self.lookbehind = self.lookbehind[0:2]
        return c, line, column


class XDRToken(abc.ABC):
    def __init__(self, line, column, value):
        self.line = line
        self.column = column
        self.value = value

    def __eq__(self, other):
        return (
            type(self) is type(other)
            and self.line == other.line
            and self.column == other.column
            and self.value == other.value
        )

    @classmethod
    @abc.abstractmethod
    def start(cls, reader):
        pass

    @classmethod
    @abc.abstractmethod
    def end(cls, reader):
        pass

    @classmethod
    def consume(cls, reader):
        c, line, col = reader.next()
        buf = c
        while True:
            if cls.end(reader):
                break
            c, _, _ = reader.next()
            buf = buf + c
        return cls(line, col, buf)

    def __repr__(self):
        return "%s{line=%d,col=%d,value={{{%s}}}}" % (
            self.__class__.__name__,
            self.line,
            self.column,
            self.value,
        )


class XDRTokenComment(XDRToken):
    @classmethod
    def start(cls, reader):
        return reader.peek() == "/" and reader.peek(skip=1) == "*"

    @classmethod
    def end(cls, reader):
        c1 = reader.last(skip=1)
        c2 = reader.last()
        if c1 == "*" and c2 == "/":
            return True

        if reader.peek() is None:
            raise Exception(
                "EOF before closing comment starting at %d:%d"
                % (reader.line, reader.column)
            )


class XDRTokenIdentifier(XDRToken):
    @classmethod
    def start(cls, reader):
        c = reader.peek()
        return c.isalpha()

    @classmethod
    def end(cls, reader):
        c = reader.peek()
        if c is None:
            return True
        return not c.isalnum() and c != "_"


class XDRTokenPunctuation(XDRToken):
    @classmethod
    def start(cls, reader):
        c = reader.peek()
        return c in [";", "=", "{", "}", ",", "[", "]", "<", ">", "*", "(", ")", ":"]

    @classmethod
    def end(cls, reader):
        return True


class XDRTokenConstant(XDRToken):
    @classmethod
    def start(cls, reader):
        c1 = reader.peek()
        c2 = reader.peek(skip=1)
        return c1.isdecimal() or (c1 == "-" and c2 is not None and c2.isdecimal())

    @classmethod
    def end(cls, reader):
        c = reader.peek()
        return (
            not c.isdecimal()
            and not c == "."
            and not c.lower() in ["x", "a", "b", "c", "d", "e", "f"]
        )


class XDRTokenCEscape(XDRToken):
    @classmethod
    def start(cls, reader):
        return reader.column == 0 and reader.peek() == "%"

    @classmethod
    def end(cls, reader):
        return reader.peek() == "\n"


class XDRTokenSpace(XDRToken):
    @classmethod
    def start(cls, reader):
        return reader.peek().isspace()

    @classmethod
    def end(cls, reader):
        c = reader.peek()
        return c is None or not c.isspace()


class XDRLexer:
    def __init__(self, fp):
        self.reader = XDRReader(fp)
        self.lookahead = []

    def _token(self):
        tokenTypes = [
            XDRTokenComment,
            XDRTokenIdentifier,
            XDRTokenCEscape,
            XDRTokenPunctuation,
            XDRTokenConstant,
            XDRTokenSpace,
        ]
        while True:
            if self.reader.peek() is None:
                return None

            for tokenType in tokenTypes:
                if tokenType.start(self.reader):
                    ret = tokenType.consume(self.reader)
                    if type(ret) not in [XDRTokenSpace, XDRTokenComment]:
                        return ret

    def next(self):
        if len(self.lookahead) > 0:
            token = self.lookahead[0]
            self.lookahead = self.lookahead[1:]
            return token
        return self._token()

    def peek(self):
        if len(self.lookahead) == 0:
            token = self._token()
            if token is None:
                return None
            self.lookahead.append(token)
        return self.lookahead[0]



# SPDX-License-Identifier: LGPL-2.1-or-later

from pathlib import Path

from rpcgen.lexer import (
    XDRLexer,
    XDRTokenIdentifier,
    XDRTokenPunctuation,
    XDRTokenConstant,
)


def test_lexer():
    p = Path(Path(__file__).parent, "simple.x")
    with p.open("r") as fp:
        lexer = XDRLexer(fp)

        tokens = []
        while True:
            tok = lexer.next()
            if tok is None:
                break
            tokens.append(tok)

    assert tokens == [
        XDRTokenIdentifier(line=3, column=0, value="const"),
        XDRTokenIdentifier(line=3, column=6, value="MAXUSERNAME"),
        XDRTokenPunctuation(line=3, column=18, value="="),
        XDRTokenConstant(line=3, column=20, value="32"),
        XDRTokenPunctuation(line=3, column=22, value=";"),
        XDRTokenIdentifier(line=4, column=0, value="const"),
        XDRTokenIdentifier(line=4, column=6, value="MAXFILELEN"),
        XDRTokenPunctuation(line=4, column=17, value="="),
        XDRTokenConstant(line=4, column=19, value="65535"),
        XDRTokenPunctuation(line=4, column=24, value=";"),
        XDRTokenIdentifier(line=5, column=0, value="const"),
        XDRTokenIdentifier(line=5, column=6, value="MAXNAMELEN"),
        XDRTokenPunctuation(line=5, column=17, value="="),
        XDRTokenConstant(line=5, column=19, value="255"),
        XDRTokenPunctuation(line=5, column=22, value=";"),
        XDRTokenIdentifier(line=10, column=0, value="enum"),
        XDRTokenIdentifier(line=10, column=5, value="filekind"),
        XDRTokenPunctuation(line=10, column=14, value="{"),
        XDRTokenIdentifier(line=11, column=3, value="TEXT"),
        XDRTokenPunctuation(line=11, column=8, value="="),
        XDRTokenConstant(line=11, column=10, value="0"),
        XDRTokenPunctuation(line=11, column=11, value=","),
        XDRTokenIdentifier(line=12, column=3, value="DATA"),
        XDRTokenPunctuation(line=12, column=8, value="="),
        XDRTokenConstant(line=12, column=10, value="1"),
        XDRTokenPunctuation(line=12, column=11, value=","),
        XDRTokenIdentifier(line=13, column=3, value="EXEC"),
        XDRTokenPunctuation(line=13, column=8, value="="),
        XDRTokenConstant(line=13, column=10, value="2"),
        XDRTokenPunctuation(line=14, column=0, value="}"),
        XDRTokenPunctuation(line=14, column=1, value=";"),
        XDRTokenIdentifier(line=19, column=0, value="union"),
        XDRTokenIdentifier(line=19, column=6, value="filetype"),
        XDRTokenIdentifier(line=19, column=15, value="switch"),
        XDRTokenPunctuation(line=19, column=22, value="("),
        XDRTokenIdentifier(line=19, column=23, value="filekind"),
        XDRTokenIdentifier(line=19, column=32, value="kind"),
        XDRTokenPunctuation(line=19, column=36, value=")"),
        XDRTokenPunctuation(line=19, column=38, value="{"),
        XDRTokenIdentifier(line=20, column=0, value="case"),
        XDRTokenIdentifier(line=20, column=5, value="TEXT"),
        XDRTokenPunctuation(line=20, column=9, value=":"),
        XDRTokenIdentifier(line=21, column=3, value="void"),
        XDRTokenPunctuation(line=21, column=7, value=";"),
        XDRTokenIdentifier(line=22, column=0, value="case"),
        XDRTokenIdentifier(line=22, column=5, value="DATA"),
        XDRTokenPunctuation(line=22, column=9, value=":"),
        XDRTokenIdentifier(line=23, column=3, value="string"),
        XDRTokenIdentifier(line=23, column=10, value="creator"),
        XDRTokenPunctuation(line=23, column=17, value="<"),
        XDRTokenIdentifier(line=23, column=18, value="MAXNAMELEN"),
        XDRTokenPunctuation(line=23, column=28, value=">"),
        XDRTokenPunctuation(line=23, column=29, value=";"),
        XDRTokenIdentifier(line=24, column=0, value="case"),
        XDRTokenIdentifier(line=24, column=5, value="EXEC"),
        XDRTokenPunctuation(line=24, column=9, value=":"),
        XDRTokenIdentifier(line=25, column=3, value="string"),
        XDRTokenIdentifier(line=25, column=10, value="interpretor"),
        XDRTokenPunctuation(line=25, column=21, value="<"),
        XDRTokenIdentifier(line=25, column=22, value="MAXNAMELEN"),
        XDRTokenPunctuation(line=25, column=32, value=">"),
        XDRTokenPunctuation(line=25, column=33, value=";"),
        XDRTokenPunctuation(line=26, column=0, value="}"),
        XDRTokenPunctuation(line=26, column=1, value=";"),
        XDRTokenIdentifier(line=30, column=0, value="struct"),
        XDRTokenIdentifier(line=30, column=7, value="file"),
        XDRTokenPunctuation(line=30, column=12, value="{"),
        XDRTokenIdentifier(line=31, column=3, value="string"),
        XDRTokenIdentifier(line=31, column=10, value="filename"),
        XDRTokenPunctuation(line=31, column=18, value="<"),
        XDRTokenIdentifier(line=31, column=19, value="MAXNAMELEN"),
        XDRTokenPunctuation(line=31, column=29, value=">"),
        XDRTokenPunctuation(line=31, column=30, value=";"),
        XDRTokenIdentifier(line=32, column=3, value="filetype"),
        XDRTokenIdentifier(line=32, column=12, value="type"),
        XDRTokenPunctuation(line=32, column=16, value=";"),
        XDRTokenIdentifier(line=33, column=3, value="string"),
        XDRTokenIdentifier(line=33, column=10, value="owner"),
        XDRTokenPunctuation(line=33, column=15, value="<"),
        XDRTokenIdentifier(line=33, column=16, value="MAXUSERNAME"),
        XDRTokenPunctuation(line=33, column=27, value=">"),
        XDRTokenPunctuation(line=33, column=28, value=";"),
        XDRTokenIdentifier(line=34, column=3, value="opaque"),
        XDRTokenIdentifier(line=34, column=10, value="data"),
        XDRTokenPunctuation(line=34, column=14, value="<"),
        XDRTokenIdentifier(line=34, column=15, value="MAXFILELEN"),
        XDRTokenPunctuation(line=34, column=25, value=">"),
        XDRTokenPunctuation(line=34, column=26, value=";"),
        XDRTokenPunctuation(line=35, column=0, value="}"),
        XDRTokenPunctuation(line=35, column=1, value=";"),
    ]



# SPDX-License-Identifier: LGPL-2.1-or-later

from pathlib import Path

from rpcgen.ast import (
    XDRSpecification,
    XDRDefinitionConstant,
    XDRDefinitionEnum,
    XDRDefinitionUnion,
    XDRDefinitionStruct,
    XDRDeclarationScalar,
    XDRDeclarationVariableArray,
    XDREnumValue,
    XDREnumBody,
    XDRStructBody,
    XDRUnionCase,
    XDRUnionBody,
    XDRTypeCustom,
    XDRTypeVoid,
    XDRTypeString,
    XDRTypeOpaque,
)
from rpcgen.parser import XDRParser


def test_parser():
    p = Path(Path(__file__).parent, "simple.x")
    with p.open("r") as fp:
        parser = XDRParser(fp)

        got = parser.parse()

    enum = XDRDefinitionEnum(
        "filekind",
        XDREnumBody(
            [
                XDREnumValue("TEXT", "0"),
                XDREnumValue("DATA", "1"),
                XDREnumValue("EXEC", "2"),
            ],
        ),
    )

    union = XDRDefinitionUnion(
        "filetype",
        XDRUnionBody(
            XDRDeclarationScalar(XDRTypeCustom("filekind", enum), "kind"),
            [
                XDRUnionCase("TEXT", XDRDeclarationScalar(XDRTypeVoid(), None)),
                XDRUnionCase(
                    "DATA",
                    XDRDeclarationVariableArray(
                        XDRTypeString(), "creator", "MAXNAMELEN"
                    ),
                ),
                XDRUnionCase(
                    "EXEC",
                    XDRDeclarationVariableArray(
                        XDRTypeString(), "interpretor", "MAXNAMELEN"
                    ),
                ),
            ],
            None,
        ),
    )

    struct = XDRDefinitionStruct(
        "file",
        XDRStructBody(
            [
                XDRDeclarationVariableArray(XDRTypeString(), "filename", "MAXNAMELEN"),
                XDRDeclarationScalar(XDRTypeCustom("filetype", union), "type"),
                XDRDeclarationVariableArray(XDRTypeString(), "owner", "MAXUSERNAME"),
                XDRDeclarationVariableArray(XDRTypeOpaque(), "data", "MAXFILELEN"),
            ]
        ),
    )

    want = XDRSpecification()
    want.definitions.extend(
        [
            XDRDefinitionConstant("MAXUSERNAME", "32"),
            XDRDefinitionConstant("MAXFILELEN", "65535"),
            XDRDefinitionConstant("MAXNAMELEN", "255"),
            enum,
            union,
            struct,
        ]
    )

    assert str(got) == str(want)



#!/usr/bin/env python3

import argparse
import copy
import os
import re
import lxml.etree

import lark


def translate_vendor(name):
    T = {
        "CPUID_VENDOR_AMD": "AMD",
        "CPUID_VENDOR_INTEL": "Intel",
        "CPUID_VENDOR_HYGON": "Hygon",
        "CPUID_VENDOR_ZHAOXIN1": "Zhaoxin CentaurHauls",
    }

    if name in T:
        return T[name]

    print(f"warning: Unknown vendor '{name}'")
    return name


def translate_feature(name, model):
    T = {
        "CPUID_6_EAX_ARAT": "arat",
        "CPUID_7_0_EBX_ADX": "adx",
        "CPUID_7_0_EBX_AVX2": "avx2",
        "CPUID_7_0_EBX_AVX512BW": "avx512bw",
        "CPUID_7_0_EBX_AVX512CD": "avx512cd",
        "CPUID_7_0_EBX_AVX512DQ": "avx512dq",
        "CPUID_7_0_EBX_AVX512ER": "avx512er",
        "CPUID_7_0_EBX_AVX512F": "avx512f",
        "CPUID_7_0_EBX_AVX512IFMA": "avx512ifma",
        "CPUID_7_0_EBX_AVX512PF": "avx512pf",
        "CPUID_7_0_EBX_AVX512VL": "avx512vl",
        "CPUID_7_0_EBX_BMI1": "bmi1",
        "CPUID_7_0_EBX_BMI2": "bmi2",
        "CPUID_7_0_EBX_CLFLUSHOPT": "clflushopt",
        "CPUID_7_0_EBX_CLWB": "clwb",
        "CPUID_7_0_EBX_ERMS": "erms",
        "CPUID_7_0_EBX_FSGSBASE": "fsgsbase",
        "CPUID_7_0_EBX_HLE": "hle",
        "CPUID_7_0_EBX_INVPCID": "invpcid",
        "CPUID_7_0_EBX_MPX": "mpx",
        "CPUID_7_0_EBX_RDSEED": "rdseed",
        "CPUID_7_0_EBX_RTM": "rtm",
        "CPUID_7_0_EBX_SHA_NI": "sha-ni",
        "CPUID_7_0_EBX_SMAP": "smap",
        "CPUID_7_0_EBX_SMEP": "smep",
        "CPUID_7_0_EBX_TSC_ADJUST": "tsc_adjust",
        "CPUID_7_0_EBX_ZERO_FCS_FDS": "zero-fcs-fds",
        "CPUID_7_0_EBX_FDP_EXCPTN_ONLY": "fdp-excptn-only",
        "CPUID_7_0_ECX_AVX512BITALG": "avx512bitalg",
        "CPUID_7_0_ECX_AVX512VNNI": "avx512vnni",
        "CPUID_7_0_ECX_AVX512_VBMI": "avx512vbmi",
        "CPUID_7_0_ECX_AVX512_VBMI2": "avx512vbmi2",
        "CPUID_7_0_ECX_AVX512_VPOPCNTDQ": "avx512-vpopcntdq",
        "CPUID_7_0_ECX_BUS_LOCK_DETECT": "bus-lock-detect",
        "CPUID_7_0_ECX_CLDEMOTE": "cldemote",
        "CPUID_7_0_ECX_GFNI": "gfni",
        "CPUID_7_0_ECX_LA57": "la57",
        "CPUID_7_0_ECX_MOVDIR64B": "movdir64b",
        "CPUID_7_0_ECX_MOVDIRI": "movdiri",
        "CPUID_7_0_ECX_PKU": "pku",
        "CPUID_7_0_ECX_RDPID": "rdpid",
        "CPUID_7_0_ECX_UMIP": "umip",
        "CPUID_7_0_ECX_VAES": "vaes",
        "CPUID_7_0_ECX_VPCLMULQDQ": "vpclmulqdq",
        "CPUID_7_0_EDX_AMX_BF16": "amx-bf16",
        "CPUID_7_0_EDX_AMX_INT8": "amx-int8",
        "CPUID_7_0_EDX_AMX_TILE": "amx-tile",
        "CPUID_7_0_EDX_ARCH_CAPABILITIES": "arch-capabilities",
        "CPUID_7_0_EDX_AVX512_4FMAPS": "avx512-4fmaps",
        "CPUID_7_0_EDX_AVX512_4VNNIW": "avx512-4vnniw",
        "CPUID_7_0_EDX_AVX512_FP16": "avx512-fp16",
        "CPUID_7_0_EDX_CORE_CAPABILITY": "core-capability",
        "CPUID_7_0_EDX_FSRM": "fsrm",
        "CPUID_7_0_EDX_SERIALIZE": "serialize",
        "CPUID_7_0_EDX_SPEC_CTRL": "spec-ctrl",
        "CPUID_7_0_EDX_SPEC_CTRL_SSBD": "ssbd",
        "CPUID_7_0_EDX_STIBP": "stibp",
        "CPUID_7_0_EDX_TSX_LDTRK": "tsx-ldtrk",
        "CPUID_7_0_EDX_AVX512_VP2INTERSECT": "avx512-vp2intersect",
        "CPUID_7_0_ECX_CET_SHSTK": "cet-ss",
        "CPUID_7_0_EDX_CET_IBT": "cet-ibt",
        "CPUID_7_0_EDX_FLUSH_L1D": "flush-l1d",
        "CPUID_7_0_EDX_MD_CLEAR": "md-clear",
        "CPUID_7_1_EAX_AMX_FP16": "amx-fp16",
        "CPUID_7_1_EAX_AVX512_BF16": "avx512-bf16",
        "CPUID_7_1_EAX_AVX_IFMA": "avx-ifma",
        "CPUID_7_1_EAX_AVX_VNNI": "avx-vnni",
        "CPUID_7_1_EAX_CMPCCXADD": "cmpccxadd",
        "CPUID_7_1_EAX_FSRC": "fsrc",
        "CPUID_7_1_EAX_FSRS": "fsrs",
        "CPUID_7_1_EAX_FZRM": "fzrm",
        "CPUID_7_1_EAX_LAM": "lam",
        "CPUID_7_1_EAX_SHA512": "sha512",
        "CPUID_7_1_EAX_SM3": "sm3",
        "CPUID_7_1_EAX_SM4": "sm4",
        "CPUID_7_1_EAX_AVX_VNNI": "avx-vnni",
        "CPUID_7_1_EAX_MOVRS": "movrs",
        "CPUID_7_1_EAX_FRED": "fred",
        "CPUID_7_1_EAX_LKGS": "lkgs",
        "CPUID_7_1_EAX_WRMSRNS": "wrmsrns",
        "CPUID_7_1_ECX_MSR_IMM": "msr-imm",
        "CPUID_7_1_EDX_AVX_NE_CONVERT": "avx-ne-convert",
        "CPUID_7_1_EDX_AVX_VNNI_INT8": "avx-vnni-int8",
        "CPUID_7_1_EDX_PREFETCHITI": "prefetchiti",
        "CPUID_7_1_EDX_AVX10": "avx10",
        "CPUID_7_1_EDX_AVX_VNNI_INT16": "avx-vnni-int16",
        "CPUID_7_1_EDX_APXF": "apxf",
        "CPUID_7_1_EDX_AMX_COMPLEX": "amx-complex",
        "CPUID_7_2_EDX_MCDT_NO": "mcdt-no",
        "CPUID_7_2_EDX_PSFD": "intel-psfd",
        "CPUID_7_2_EDX_IPRED_CTRL": "ipred-ctrl",
        "CPUID_7_2_EDX_RRSBA_CTRL": "rrsba-ctrl",
        "CPUID_7_2_EDX_BHI_CTRL": "bhi-ctrl",
        "CPUID_7_2_EDX_DDPD_U": "ddpd-u",
        "CPUID_1E_1_EAX_AMX_AVX512": "amx-avx512",
        "CPUID_1E_1_EAX_AMX_BF16_ALIAS": "amx-bf16-alias",
        "CPUID_1E_1_EAX_AMX_COMPLEX_ALIAS": "amx-complex-alias",
        "CPUID_1E_1_EAX_AMX_FP16_ALIAS": "amx-fp16-alias",
        "CPUID_1E_1_EAX_AMX_FP8": "amx-fp8",
        "CPUID_1E_1_EAX_AMX_INT8_ALIAS": "amx-int8-alias",
        "CPUID_1E_1_EAX_AMX_MOVRS": "amx-movrs",
        "CPUID_1E_1_EAX_AMX_TF32": "amx-tf32",
        "CPUID_24_0_EBX_AVX10_128": "avx10-128",
        "CPUID_24_0_EBX_AVX10_256": "avx10-256",
        "CPUID_24_0_EBX_AVX10_512": "avx10-512",
        "CPUID_24_1_ECX_AVX10_VNNI_INT": "avx10-vnni-int",
        "CPUID_29_0_EBX_APX_NCI_NDD_NF": "apx-nci-ndd-nf",
        "CPUID_APM_INVTSC": "invtsc",
        "CPUID_8000_0007_EBX_OVERFLOW_RECOV": "overflow-recov",
        "CPUID_8000_0007_EBX_SUCCOR": "succor",
        "CPUID_8000_0008_EBX_AMD_PSFD": "amd-psfd",
        "CPUID_8000_0008_EBX_AMD_SSBD": "amd-ssbd",
        "CPUID_8000_0008_EBX_CLZERO": "clzero",
        "CPUID_8000_0008_EBX_IBPB": "ibpb",
        "CPUID_8000_0008_EBX_IBRS": "ibrs",
        "CPUID_8000_0008_EBX_STIBP": "amd-stibp",
        "CPUID_8000_0008_EBX_STIBP_ALWAYS_ON": "stibp-always-on",
        "CPUID_8000_0008_EBX_WBNOINVD": "wbnoinvd",
        "CPUID_8000_0008_EBX_XSAVEERPTR": "xsaveerptr",
        "CPUID_8000_0021_EAX_AUTO_IBRS": "auto-ibrs",
        "CPUID_8000_0021_EAX_LFENCE_ALWAYS_SERIALIZING":
            "lfence-always-serializing",
        "CPUID_8000_0021_EAX_NULL_SEL_CLR_BASE": "null-sel-clr-base",
        "CPUID_8000_0021_EAX_NO_NESTED_DATA_BP": "no-nested-data-bp",
        "CPUID_8000_0021_EAX_FS_GS_BASE_NS": "fs-gs-base-ns",
        "CPUID_8000_0021_EAX_PREFETCHI": "prefetchi",
        "CPUID_8000_0021_EAX_VERW_CLEAR": "verw-clear",
        "CPUID_8000_0021_ECX_TSA_SQ_NO": "tsa-sq-no",
        "CPUID_8000_0021_ECX_TSA_L1_NO": "tsa-l1-no",
        "CPUID_8000_0021_EAX_SBPB": "sbpb",
        "CPUID_8000_0021_EAX_IBPB_BRTYPE": "ibpb-brtype",
        "CPUID_8000_0021_EAX_SRSO_USER_KERNEL_NO": "srso-user-kernel-no",
        "CPUID_8000_0022_EAX_PERFMON_V2": "perfmon-v2",
        "CPUID_C000_0001_EDX_XSTORE": "xstore",
        "CPUID_C000_0001_EDX_XSTORE_EN": "xstore-en",
        "CPUID_C000_0001_EDX_XCRYPT": "xcrypt",
        "CPUID_C000_0001_EDX_XCRYPT_EN": "xcrypt-en",
        "CPUID_C000_0001_EDX_ACE2": "ace2",
        "CPUID_C000_0001_EDX_ACE2_EN": "ace2-en",
        "CPUID_C000_0001_EDX_PHE": "phe",
        "CPUID_C000_0001_EDX_PHE_EN": "phe-en",
        "CPUID_C000_0001_EDX_PMM": "pmm",
        "CPUID_C000_0001_EDX_PMM_EN": "pmm-en",
        "CPUID_ACPI": "acpi",
        "CPUID_APIC": "apic",
        "CPUID_CLFLUSH": "clflush",
        "CPUID_CMOV": "cmov",
        "CPUID_CX8": "cx8",
        "CPUID_DE": "de",
        "CPUID_D_1_EAX_XFD": "xfd",
        "CPUID_EXT2_3DNOW": "3dnow",
        "CPUID_EXT2_3DNOWEXT": "3dnowext",
        "CPUID_EXT2_FFXSR": "fxsr_opt",
        "CPUID_EXT2_LM": "lm",
        "CPUID_EXT2_MMXEXT": "mmxext",
        "CPUID_EXT2_NX": "nx",
        "CPUID_EXT2_PDPE1GB": "pdpe1gb",
        "CPUID_EXT2_RDTSCP": "rdtscp",
        "CPUID_EXT2_SYSCALL": "syscall",
        "CPUID_EXT3_3DNOWPREFETCH": "3dnowprefetch",
        "CPUID_EXT3_ABM": "abm",
        "CPUID_EXT3_CR8LEG": "cr8legacy",
        "CPUID_EXT3_FMA4": "fma4",
        "CPUID_EXT3_LAHF_LM": "lahf_lm",
        "CPUID_EXT3_MISALIGNSSE": "misalignsse",
        "CPUID_EXT3_OSVW": "osvw",
        "CPUID_EXT3_PERFCORE": "perfctr_core",
        "CPUID_EXT3_SSE4A": "sse4a",
        "CPUID_EXT3_SVM": "svm",
        "CPUID_EXT3_TBM": "tbm",
        "CPUID_EXT3_XOP": "xop",
        "CPUID_EXT_AES": "aes",
        "CPUID_EXT_AVX": "avx",
        "CPUID_EXT_CX16": "cx16",
        "CPUID_EXT_F16C": "f16c",
        "CPUID_EXT_FMA": "fma",
        "CPUID_EXT_MOVBE": "movbe",
        "CPUID_EXT_PCID": "pcid",
        "CPUID_EXT_PCLMULQDQ": "pclmuldq",
        "CPUID_EXT_POPCNT": "popcnt",
        "CPUID_EXT_RDRAND": "rdrand",
        "CPUID_EXT_SSE3": "pni",
        "CPUID_EXT_SSE41": "sse4.1",
        "CPUID_EXT_SSE42": "sse4.2",
        "CPUID_EXT_SSSE3": "ssse3",
        "CPUID_EXT_TSC_DEADLINE_TIMER": "tsc-deadline",
        "CPUID_EXT_X2APIC": "x2apic",
        "CPUID_EXT_XSAVE": "xsave",
        "CPUID_FP87": "fpu",
        "CPUID_FXSR": "fxsr",
        "CPUID_MCA": "mca",
        "CPUID_MCE": "mce",
        "CPUID_MMX": "mmx",
        "CPUID_MSR": "msr",
        "CPUID_MTRR": "mtrr",
        "CPUID_PAE": "pae",
        "CPUID_PAT": "pat",
        "CPUID_PGE": "pge",
        "CPUID_PSE": "pse",
        "CPUID_PSE36": "pse36",
        "CPUID_SEP": "sep",
        "CPUID_SS": "ss",
        "CPUID_SSE": "sse",
        "CPUID_SSE2": "sse2",
        "CPUID_SVM_NPT": "npt",
        "CPUID_SVM_NRIPSAVE": "nrip-save",
        "CPUID_SVM_SVME_ADDR_CHK": "svme-addr-chk",
        "CPUID_SVM_VNMI": "vnmi",
        "CPUID_SVM_LBRV": "lbrv",
        "CPUID_SVM_TSCSCALE": "tsc-scale",
        "CPUID_SVM_VMCBCLEAN": "vmcb-clean",
        "CPUID_SVM_FLUSHASID": "flushbyasid",
        "CPUID_SVM_PAUSEFILTER": "pause-filter",
        "CPUID_SVM_PFTHRESHOLD": "pfthreshold",
        "CPUID_SVM_V_VMSAVE_VMLOAD": "v-vmsave-vmload",
        "CPUID_SVM_VGIF": "vgif",
        "CPUID_SVM_GMET": "gmet",
        "CPUID_TSC": "tsc",
        "CPUID_VME": "vme",
        "CPUID_XSAVE_XGETBV1": "xgetbv1",
        "CPUID_XSAVE_XSAVEC": "xsavec",
        "CPUID_XSAVE_XSAVEOPT": "xsaveopt",
        "CPUID_XSAVE_XSAVES": "xsaves",
        "MSR_ARCH_CAP_FBSDP_NO": "fbsdp-no",
        "MSR_ARCH_CAP_IBRS_ALL": "ibrs-all",
        "MSR_ARCH_CAP_MDS_NO": "mds-no",
        "MSR_ARCH_CAP_PBRSB_NO": "pbrsb-no",
        "MSR_ARCH_CAP_PSCHANGE_MC_NO": "pschange-mc-no",
        "MSR_ARCH_CAP_PSDP_NO": "psdp-no",
        "MSR_ARCH_CAP_RDCL_NO": "rdctl-no",
        "MSR_ARCH_CAP_SBDR_SSDP_NO": "sbdr-ssdp-no",
        "MSR_ARCH_CAP_SKIP_L1DFL_VMENTRY": "skip-l1dfl-vmentry",
        "MSR_ARCH_CAP_TAA_NO": "taa-no",
        "MSR_ARCH_CAP_BHI_NO": "bhi-no",
        "MSR_CORE_CAP_SPLIT_LOCK_DETECT": "split-lock-detect",
        "MSR_ARCH_CAP_GDS_NO": "gds-no",
        "MSR_ARCH_CAP_RFDS_NO": "rfds-no",
        "MSR_ARCH_CAP_SSB_NO": "ssb-no",

        # FEAT_VMX_PROCBASED_CTLS
        "VMX_CPU_BASED_VIRTUAL_INTR_PENDING": "vmx-vintr-pending",
        "VMX_CPU_BASED_USE_TSC_OFFSETING": "vmx-tsc-offset",
        "VMX_CPU_BASED_HLT_EXITING": "vmx-hlt-exit",
        "VMX_CPU_BASED_INVLPG_EXITING": "vmx-invlpg-exit",
        "VMX_CPU_BASED_MWAIT_EXITING": "vmx-mwait-exit",
        "VMX_CPU_BASED_RDPMC_EXITING": "vmx-rdpmc-exit",
        "VMX_CPU_BASED_RDTSC_EXITING": "vmx-rdtsc-exit",
        "VMX_CPU_BASED_CR3_LOAD_EXITING": "vmx-cr3-load-noexit",
        "VMX_CPU_BASED_CR3_STORE_EXITING": "vmx-cr3-store-noexit",
        "VMX_CPU_BASED_CR8_LOAD_EXITING": "vmx-cr8-load-exit",
        "VMX_CPU_BASED_CR8_STORE_EXITING": "vmx-cr8-store-exit",
        "VMX_CPU_BASED_TPR_SHADOW": "vmx-flexpriority",
        "VMX_CPU_BASED_VIRTUAL_NMI_PENDING": "vmx-vnmi-pending",
        "VMX_CPU_BASED_MOV_DR_EXITING": "vmx-movdr-exit",
        "VMX_CPU_BASED_UNCOND_IO_EXITING": "vmx-io-exit",
        "VMX_CPU_BASED_USE_IO_BITMAPS": "vmx-io-bitmap",
        "VMX_CPU_BASED_MONITOR_TRAP_FLAG": "vmx-mtf",
        "VMX_CPU_BASED_USE_MSR_BITMAPS": "vmx-msr-bitmap",
        "VMX_CPU_BASED_MONITOR_EXITING": "vmx-monitor-exit",
        "VMX_CPU_BASED_PAUSE_EXITING": "vmx-pause-exit",
        "VMX_CPU_BASED_ACTIVATE_SECONDARY_CONTROLS": "vmx-secondary-ctls",

        # FEAT_VMX_SECONDARY_CTLS
        "VMX_SECONDARY_EXEC_VIRTUALIZE_APIC_ACCESSES": "vmx-apicv-xapic",
        "VMX_SECONDARY_EXEC_ENABLE_EPT": "vmx-ept",
        "VMX_SECONDARY_EXEC_DESC": "vmx-desc-exit",
        "VMX_SECONDARY_EXEC_RDTSCP": "vmx-rdtscp-exit",
        "VMX_SECONDARY_EXEC_VIRTUALIZE_X2APIC_MODE": "vmx-apicv-x2apic",
        "VMX_SECONDARY_EXEC_ENABLE_VPID": "vmx-vpid",
        "VMX_SECONDARY_EXEC_WBINVD_EXITING": "vmx-wbinvd-exit",
        "VMX_SECONDARY_EXEC_UNRESTRICTED_GUEST": "vmx-unrestricted-guest",
        "VMX_SECONDARY_EXEC_APIC_REGISTER_VIRT": "vmx-apicv-register",
        "VMX_SECONDARY_EXEC_VIRTUAL_INTR_DELIVERY": "vmx-apicv-vid",
        "VMX_SECONDARY_EXEC_PAUSE_LOOP_EXITING": "vmx-ple",
        "VMX_SECONDARY_EXEC_RDRAND_EXITING": "vmx-rdrand-exit",
        "VMX_SECONDARY_EXEC_ENABLE_INVPCID": "vmx-invpcid-exit",
        "VMX_SECONDARY_EXEC_ENABLE_VMFUNC": "vmx-vmfunc",
        "VMX_SECONDARY_EXEC_SHADOW_VMCS": "vmx-shadow-vmcs",
        "VMX_SECONDARY_EXEC_ENCLS_EXITING": "vmx-encls-exit",
        "VMX_SECONDARY_EXEC_RDSEED_EXITING": "vmx-rdseed-exit",
        "VMX_SECONDARY_EXEC_ENABLE_PML": "vmx-pml",
        "VMX_SECONDARY_EXEC_XSAVES": "vmx-xsaves",
        "VMX_SECONDARY_EXEC_TSC_SCALING": "vmx-tsc-scaling",
        "VMX_SECONDARY_EXEC_ENABLE_USER_WAIT_PAUSE": "vmx-enable-user-wait-pause",
        "VMX_SECONDARY_EXEC_MODE_BASED_EPT_EXEC": "vmx-mbec",

        # FEAT_VMX_PINBASED_CTLS
        "VMX_PIN_BASED_EXT_INTR_MASK": "vmx-intr-exit",
        "VMX_PIN_BASED_NMI_EXITING": "vmx-nmi-exit",
        "VMX_PIN_BASED_VIRTUAL_NMIS": "vmx-vnmi",
        "VMX_PIN_BASED_VMX_PREEMPTION_TIMER": "vmx-preemption-timer",
        "VMX_PIN_BASED_POSTED_INTR": "vmx-posted-intr",

        # FEAT_VMX_EXIT_CTLS
        "VMX_VM_EXIT_SAVE_DEBUG_CONTROLS": "vmx-exit-nosave-debugctl",
        "VMX_VM_EXIT_LOAD_IA32_PERF_GLOBAL_CTRL": "vmx-exit-load-perf-global-ctrl",
        "VMX_VM_EXIT_ACK_INTR_ON_EXIT": "vmx-exit-ack-intr",
        "VMX_VM_EXIT_SAVE_IA32_PAT": "vmx-exit-save-pat",
        "VMX_VM_EXIT_LOAD_IA32_PAT": "vmx-exit-load-pat",
        "VMX_VM_EXIT_SAVE_IA32_EFER": "vmx-exit-save-efer",
        "VMX_VM_EXIT_LOAD_IA32_EFER": "vmx-exit-load-efer",
        "VMX_VM_EXIT_SAVE_VMX_PREEMPTION_TIMER": "vmx-exit-save-preemption-timer",
        "VMX_VM_EXIT_CLEAR_BNDCFGS": "vmx-exit-clear-bndcfgs",
        "VMX_VM_EXIT_CLEAR_IA32_RTIT_CTL": "vmx-exit-clear-rtit-ctl",
        "VMX_VM_EXIT_LOAD_IA32_PKRS": "vmx-exit-load-pkrs",
        "VMX_VM_EXIT_SAVE_CET": "vmx-exit-save-cet",
        "VMX_VM_EXIT_ACTIVATE_SECONDARY_CONTROLS": "vmx-exit-secondary-ctls",

        # FEAT_VMX_ENTRY_CTLS
        "VMX_VM_ENTRY_LOAD_DEBUG_CONTROLS": "vmx-entry-noload-debugctl",
        "VMX_VM_ENTRY_IA32E_MODE": "vmx-entry-ia32e-mode",
        "VMX_VM_ENTRY_LOAD_IA32_PERF_GLOBAL_CTRL": "vmx-entry-load-perf-global-ctrl",
        "VMX_VM_ENTRY_LOAD_IA32_PAT": "vmx-entry-load-pat",
        "VMX_VM_ENTRY_LOAD_IA32_EFER": "vmx-entry-load-efer",
        "VMX_VM_ENTRY_LOAD_BNDCFGS": "vmx-entry-load-bndcfgs",
        "VMX_VM_ENTRY_LOAD_IA32_RTIT_CTL": "vmx-entry-load-rtit-ctl",
        "VMX_VM_ENTRY_LOAD_IA32_PKRS": "vmx-entry-load-pkrs",
        "VMX_VM_ENTRY_LOAD_CET": "vmx-entry-load-cet",
        "VMX_VM_ENTRY_LOAD_IA32_FRED": "vmx-entry-load-fred",

        # FEAT_VMX_MISC
        "MSR_VMX_MISC_STORE_LMA": "vmx-store-lma",
        "MSR_VMX_MISC_ACTIVITY_HLT": "vmx-activity-hlt",
        "MSR_VMX_MISC_ACTIVITY_SHUTDOWN": "vmx-activity-shutdown",
        "MSR_VMX_MISC_ACTIVITY_WAIT_SIPI": "vmx-activity-wait-sipi",
        "MSR_VMX_MISC_VMWRITE_VMEXIT": "vmx-vmwrite-vmexit-fields",
        "MSR_VMX_MISC_ZERO_LEN_INJECT": "vmx-zero-len-inject",

        # FEAT_VMX_EPT_VPID_CAPS
        "MSR_VMX_EPT_EXECONLY": "vmx-ept-execonly",
        "MSR_VMX_EPT_PAGE_WALK_LENGTH_4": "vmx-page-walk-4",
        "MSR_VMX_EPT_PAGE_WALK_LENGTH_5": "vmx-page-walk-5",
        "MSR_VMX_EPT_2MB": "vmx-ept-2mb",
        "MSR_VMX_EPT_1GB": "vmx-ept-1gb",
        "MSR_VMX_EPT_INVEPT": "vmx-invept",
        "MSR_VMX_EPT_AD_BITS": "vmx-eptad",
        "MSR_VMX_EPT_ADVANCED_VMEXIT_INFO": "vmx-ept-advanced-exitinfo",
        "MSR_VMX_EPT_INVEPT_SINGLE_CONTEXT": "vmx-invept-single-context",
        "MSR_VMX_EPT_INVEPT_ALL_CONTEXT": "vmx-invept-all-context",
        "MSR_VMX_EPT_INVVPID": "vmx-invvpid",
        "MSR_VMX_EPT_INVVPID_SINGLE_ADDR": "vmx-invvpid-single-addr",
        "MSR_VMX_EPT_INVVPID_ALL_CONTEXT": "vmx-invvpid-all-context",
        "MSR_VMX_EPT_INVVPID_SINGLE_CONTEXT_NOGLOBALS": "vmx-invvpid-single-context-noglobals",

        # FEAT_VMX_BASIC
        "MSR_VMX_BASIC_INS_OUTS": "vmx-ins-outs",
        "MSR_VMX_BASIC_TRUE_CTLS": "vmx-true-ctls",
        "MSR_VMX_BASIC_ANY_ERRCODE": "vmx-any-errcode",
        "MSR_VMX_BASIC_NESTED_EXCEPTION": "vmx-nested-exception",

        # FEAT_VMX_VMFUNC
        "MSR_VMX_VMFUNC_EPT_SWITCHING": "vmx-eptp-switching",
    }

    ignore = any([
        name in ("0", "model", "model-id", "stepping"),
        name in ("CPUID_EXT_MONITOR", "monitor"),
        name in ("MSR_VMX_BASIC_DUAL_MONITOR", "dual-monitor"),
        name in ("CPUID_EXT3_TOPOEXT", "topoext"),
        name in ("MSR_VMX_EPT_UC", "MSR_VMX_EPT_WB"),
        name in ("MSR_VMX_EPT_INVVPID_SINGLE_CONTEXT"),
        name in ("x-force-cpuid-0x1f", "avx10-version"),
        name in ("VMX_VM_EXIT_HOST_ADDR_SPACE_SIZE"),
    ])

    if ignore:
        return None

    if name in T:
        return T[name]

    for v in T.values():
        if name.replace("-", "_") == v.replace("-", "_"):
            return v

    print(f"warning: Unknown feature '{name}' in model '{model}'")
    return name


def readline_cont(f):
    """Read one logical line from a file `f` i.e. continues lines that end in
    a backslash."""

    line = f.readline()
    while line.endswith("\\\n"):
        line = line[:-2] + " " + f.readline()
    return line


def read_builtin_x86_defs(filename):
    """Extract content between begin_mark and end_mark from file `filename` as
    string, while expanding shorthand macros like "I486_FEATURES"."""

    begin_mark = re.compile(
        "^static( const)? X86CPUDefinition builtin_x86_defs\\[\\] = {$")
    end_mark = "};\n"
    shorthand = re.compile("^#define ([A-Z0-9_]+_FEATURES) (.*)$")
    lines = list()
    shorthands = dict()

    with open(filename, "rt") as f:
        while True:
            line = readline_cont(f)
            if not line:
                raise RuntimeError("begin mark not found")
            match = begin_mark.match(line)
            if match:
                break
            match = shorthand.match(line)
            if match:
                # TCG definitions are irrelevant for cpu models
                newk = match.group(1)
                if newk.startswith("TCG_"):
                    continue

                # remove comments, whitespace and bit operators, effectively
                # turning the bitfield into a list
                newv = re.sub("([()|\t\n])|(/\\*.*?\\*/)", " ", match.group(2))

                # resolve recursive shorthands
                for k, v in shorthands.items():
                    newv = newv.replace(k, v)

                shorthands[newk] = newv

        while True:
            line = readline_cont(f)
            if line == end_mark:
                break
            if not line:
                raise RuntimeError("end marker not found")

            # apply shorthands
            for k, v in shorthands.items():
                line = line.replace(k, v)
            lines.append(line)

    return "".join(lines)


def transform(item):
    """Recursively transform a Lark syntax tree into python native objects."""

    if isinstance(item, lark.lexer.Token):
        return str(item)

    if item.data == "list":
        retval = list()
        for child in item.children:
            value = transform(child)
            if value is None:
                continue
            retval.append(value)
        return retval

    if item.data == "map":
        retval = dict()
        for child in item.children:
            if len(child.children) != 2:
                raise RuntimeError("map entry with more than 2 elements")
            key = transform(child.children[0])
            value = transform(child.children[1])
            if key is None:
                raise RuntimeError("map entry with 'None' key")
            if value is None:
                continue
            retval[key] = value
        return retval

    if item.data == "text":
        retval = list()
        for child in item.children:
            value = transform(child)
            if value is None:
                continue
            retval.append(value)
        return " ".join(retval)

    if item.data == "value":
        if item.children:
            raise RuntimeError("empty list is not empty")
        return None

    raise RuntimeError("unexpected item type")


def get_signature(outdir, model):
    file = os.path.join(outdir, f"x86_{model}.xml")

    if not os.path.isfile(file):
        return None

    xml = lxml.etree.parse(file)

    signature = []
    for sig in xml.xpath("//signature"):
        attr = sig.attrib
        family = attr["family"]
        model = attr["model"]
        if "stepping" in attr:
            stepping = attr["stepping"]
        else:
            stepping = None

        signature.append((family, model, stepping))

    return signature


def expand_model(outdir, model):
    """Expand a qemu cpu model description that has its feature split up into
    different fields and may have differing versions into several libvirt-
    friendly cpu models."""

    result = {
        "name": model.pop(".name"),
        "vendor": translate_vendor(model.pop(".vendor")),
        "features": set(),
        "extra": dict(),
        "signature": list(),
    }

    if ".family" in model and ".model" in model:
        result["signature"].append((str(int(model.pop(".family"), base=0)),
                                    str(int(model.pop(".model"), base=0)),
                                    None))

    for k in [k for k in model if k.startswith(".features")]:
        v = model.pop(k)
        for feature in v.split():
            translated = translate_feature(feature, result["name"])
            if translated:
                result["features"].add(translated)

    versions = model.pop(".versions", [])
    for k, v in model.items():
        result["extra"]["model" + k] = v

    print(result['name'])
    yield result

    if len(versions) == 0:
        versions.append({".version": "1"})

    name = result["name"]
    for version in versions:
        result = copy.deepcopy(result)

        ver = int(version.pop(".version"))
        result["name"] = f"{name}-v{ver}"
        result["base"] = name

        alias = version.pop(".alias", None)
        if not alias and ver == 1:
            alias = name

            sig = get_signature(outdir, name)
            if sig:
                result["signature"] = sig

        props = version.pop(".props", dict())
        for k, v in props:
            if k not in ("model-id", "stepping", "model"):
                k = translate_feature(k, result["name"])
            if k is None:
                continue

            if v == "on":
                result["features"].add(k)
            elif v == "off" and k in result["features"]:
                result["features"].remove(k)
            else:
                result["extra"]["property." + k] = v

        for k, v in version.items():
            result["extra"]["version" + k] = v

        if alias:
            print(f"v{ver}: {result['name']} => {alias}")
            yield {
                "vendor": result["vendor"],
                "name": result["name"],
                "base": result["base"],
                "alias": alias,
                "extra": None,
                "features": [],
            }

            if ver != 1:
                result["name"] = alias
                print(f"v{ver}: {result['name']}")
                yield result
        else:
            print(f"v{ver}: {result['name']}")
            yield result


def output_model(f, extra, model):
    if model["extra"]:
        with open(extra, "wt") as ex:
            ex.write("# THIS FILE SHOULD NEVER BE ADDED TO A COMMIT\n")
            ex.write("extra info from qemu:\n")
            for k, v in model["extra"].items():
                ex.write(f"  {k}: {v}\n")

    decode = "off" if "base" in model else "on"

    f.write("<cpus>\n")
    f.write(f"  <model name='{model['name']}'>\n")
    f.write(f"    <decode host='on' guest='{decode}'/>\n")

    if "alias" in model:
        f.write(f"    <model name='{model['alias']}'/>\n")
    else:
        for sig_family, sig_model, sig_stepping in model['signature']:
            f.write(f"    <signature family='{sig_family}' model='{sig_model}'")
            if sig_stepping:
                f.write(f" stepping='{sig_stepping}'")
            f.write("/>\n")
        f.write(f"    <vendor name='{model['vendor']}'/>\n")

    for feature in sorted(model["features"]):
        f.write(f"    <feature name='{feature}'/>\n")
    f.write("  </model>\n")
    f.write("</cpus>\n")


def update_index(outdir, models):
    index = os.path.join(outdir, "index.xml")
    xml = lxml.etree.parse(index)

    for vendor, files in models.items():
        groups = xml.xpath(f"//arch[@name='x86']/group[@name='{vendor} CPU models']")
        if not groups:
            continue

        group = groups[-1]
        children = group.getchildren()
        if children:
            last = children[-1]
            group_indent = last.tail
            indent = f"{group_indent}  "
            last.tail = indent
        else:
            group_indent = f"{group.tail}  "
            indent = f"{group_indent}  "
            group.text = f"{group_indent}  "

        for file in files:
            include = lxml.etree.SubElement(group, "include", filename=file)
            include.tail = indent

        group.getchildren()[-1].tail = group_indent

    out = lxml.etree.tostring(xml, encoding="UTF-8")
    out = out.decode("UTF-8").replace('"', "'")

    with open(index, "w") as f:
        f.write(out)
        f.write("\n")


def update_meson(outdir, models):
    meson = os.path.join(outdir, "meson.build")

    with open(meson, "r") as f:
        lines = f.readlines()

    start = None
    end = None
    for i in range(len(lines)):
        if start is None and lines[i].startswith("cpumap_data ="):
            start = i + 1

        if start is not None and lines[i] == "]\n":
            end = i
            break

    xmls = lines[start:end]
    for files in models.values():
        xmls.extend([f"  '{file}',\n" for file in files])

    with open(meson, "w") as f:
        f.writelines(lines[:start])
        f.writelines(sorted(xmls, key=str.lower))
        f.writelines(lines[end:])


def main():
    parser = argparse.ArgumentParser(
        description="Synchronize x86 cpu models from QEMU i386 target.")
    parser.add_argument(
        "qemu",
        help="Path to QEMU source code",
        type=os.path.realpath)
    parser.add_argument(
        "outdir",
        help="Path to 'src/cpu_map' directory in the libvirt repository",
        type=os.path.realpath)

    args = parser.parse_args()

    cpufile = os.path.join(args.qemu, 'target/i386/cpu.c')
    if not os.path.isfile(cpufile):
        parser.print_help()
        exit("QEMU source directory not found")

    builtin_x86_defs = read_builtin_x86_defs(cpufile)

    ast = lark.Lark(r"""
        list: value ( "," value )* ","?
        map: keyvalue ( "," keyvalue )* ","?
        keyvalue: IDENTIFIER "=" value
        ?value: text | "{" "}" | "{" list "}" | "{" map "}"
        text: (IDENTIFIER | "\"" (/[^"]+/)? "\"")+
        IDENTIFIER: /[\[\]\._&a-zA-Z0-9]/+
        %ignore (" " | "\r" | "\n" | "\t" | "|" )+
        %ignore "(" ( "X86CPUVersionDefinition" | "PropValue" ) "[])"
        %ignore "//" /.*?/ "\n"
        %ignore "/*" /(.|\n)*?/ "*/"
        """, start="list").parse(builtin_x86_defs)

    models_json = transform(ast)

    models = list()
    for model in models_json:
        models.extend(expand_model(args.outdir, model))

    files = dict()

    for model in models:
        name = f"x86_{model['name']}.xml"
        path = os.path.join(args.outdir, name)

        if os.path.isfile(path):
            # Ignore existing models as CPU models in libvirt should never
            # change once released.
            continue

        vendor = model['vendor']
        if vendor:
            if vendor not in files:
                files[vendor] = []
            files[vendor].append(name)

        extra = os.path.join(args.outdir, f"x86_{model['name']}.extra")
        with open(path, "wt") as f:
            output_model(f, extra, model)

    update_index(args.outdir, files)
    update_meson(args.outdir, files)

    features = set()
    for model in models:
        features.update(model["features"])

    try:
        filename = os.path.join(args.outdir, "x86_features.xml")
        dom = lxml.etree.parse(filename)
        known = [x.attrib["name"] for x in dom.getroot().iter("feature")]
        unknown = [x for x in features if x not in known and x is not None]
    except Exception as e:
        unknown = []
        print(f"warning: Unable to read libvirt x86_features.xml: {e}")

    for x in unknown:
        print(f"warning: Feature unknown to libvirt: {x}")


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: LGPL-2.1-or-later

import abc


class XDRSpecification:
    def __init__(self):
        self.definitions = []

    def __repr__(self):
        return "\n".join([repr(a) for a in self.definitions])


class XDRDefinition(abc.ABC):
    pass


class XDRDefinitionConstant(XDRDefinition):
    def __init__(self, name, value):
        self.name = name
        self.value = value

    def __repr__(self):
        return "const:{name=%s,value=%s}" % (self.name, self.value)


class XDRDefinitionTypedef(XDRDefinition):
    def __init__(self, decl):
        self.decl = decl

    def __repr__(self):
        return "typedef:{decl=%s}" % (self.decl)


class XDRDefinitionEnum(XDRDefinition):
    def __init__(self, name, body):
        self.name = name
        self.body = body

    def __repr__(self):
        return "enum:{name=%s,body=%s}" % (self.name, self.body)


class XDRDefinitionStruct(XDRDefinition):
    def __init__(self, name, body):
        self.name = name
        self.body = body

    def __repr__(self):
        return "struct:{name=%s,body=%s}" % (self.name, self.body)


class XDRDefinitionUnion(XDRDefinition):
    def __init__(self, name, body):
        self.name = name
        self.body = body

    def __repr__(self):
        return "union:{name=%s,body=%s}" % (self.name, self.body)


class XDRDefinitionCEscape(XDRDefinition):
    def __init__(self, code):
        self.code = code

    def __repr__(self):
        return "c-escape:{code=%s}" % (self.code)


class XDRDeclaration(abc.ABC):
    def __init__(self, typ, identifier):
        self.typ = typ
        self.identifier = identifier


class XDRDeclarationScalar(XDRDeclaration):
    def __repr__(self):
        return "scalar:{type=%s,identifier=%s}" % (self.typ, self.identifier)


class XDRDeclarationPointer(XDRDeclaration):
    def __repr__(self):
        return "pointer:{type=%s,identifier=%s}" % (self.typ, self.identifier)


class XDRDeclarationFixedArray(XDRDeclaration):
    def __init__(self, typ, identifier, length):
        super().__init__(typ, identifier)
        self.length = length

    def __repr__(self):
        return "fixed-array:{type=%s,identifier=%s,length=%s}" % (
            self.typ,
            self.identifier,
            self.length,
        )


class XDRDeclarationVariableArray(XDRDeclaration):
    def __init__(self, typ, identifier, maxlength):
        super().__init__(typ, identifier)
        self.maxlength = maxlength

    def __repr__(self):
        return "variable-array:{type=%s,identifier=%s,maxlength=%s}" % (
            self.typ,
            self.identifier,
            self.maxlength,
        )


class XDRType(abc.ABC):
    def __repr__(self):
        name = self.__class__.__name__
        return name[7:].lower()

    def is_scalar(self):
        return False


class XDRTypeScalar(XDRType):
    def is_scalar(self):
        return True


class XDRTypeVoid(XDRTypeScalar):
    pass


class XDRTypeChar(XDRTypeScalar):
    pass


class XDRTypeUnsignedChar(XDRTypeScalar):
    pass


class XDRTypeShort(XDRTypeScalar):
    pass


class XDRTypeUnsignedShort(XDRTypeScalar):
    pass


class XDRTypeInt(XDRTypeScalar):
    pass


class XDRTypeUnsignedInt(XDRTypeScalar):
    pass


class XDRTypeHyper(XDRTypeScalar):
    pass


class XDRTypeUnsignedHyper(XDRTypeScalar):
    pass


class XDRTypeFloat(XDRTypeScalar):
    pass


class XDRTypeDouble(XDRTypeScalar):
    pass


class XDRTypeBool(XDRTypeScalar):
    pass


class XDRTypeOpaque(XDRType):
    pass


class XDRTypeString(XDRType):
    pass


class XDRTypeCustom(XDRType):
    def __init__(self, identifier, definition):
        self.identifier = identifier
        self.definition = definition

    def is_scalar(self):
        if type(self.definition) is XDRDefinitionEnum:
            return True
        if type(self.definition) is XDRDefinitionTypedef:
            return self.definition.decl.typ.is_scalar()
        return False

    def __repr__(self):
        return "custom{identifier=%s,definition=%s}" % (
            self.identifier,
            self.definition,
        )


class XDREnumValue:
    def __init__(self, name, value):
        self.name = name
        self.value = value

    def __repr__(self):
        return "%s=%s" % (self.name, self.value)


class XDREnumBody:
    def __init__(self, values):
        self.values = values

    def __repr__(self):
        return "enum-body{values=[" + ",".join([str(a) for a in self.values]) + "]}"


class XDRTypeEnum(XDRTypeScalar):
    def __init__(self, body):
        self.body = body

    def __repr__(self):
        return "enum{%s}" % self.body


class XDRStructBody:
    def __init__(self, fields):
        self.fields = fields

    def __repr__(self):
        return "struct-body{fields=[" + ",".join([str(a) for a in self.fields]) + "]}"


class XDRTypeStruct(XDRType):
    def __init__(self, body):
        self.body = body

    def __repr__(self):
        return "struct{%s}" % self.body


class XDRUnionCase:
    def __init__(self, value, decl):
        self.value = value
        self.decl = decl

    def __repr__(self):
        return "%s=%s" % (self.value, self.decl)


class XDRUnionBody:
    def __init__(self, discriminator, cases, default):
        self.discriminator = discriminator
        self.cases = cases
        self.default = default

    def __repr__(self):
        return (
            "union-body{discriminator=%s,cases=["
            + ",".join([str(a) for a in self.cases])
            + "],default=%s}"
        ) % (self.discriminator, self.default)


class XDRTypeUnion(XDRType):
    def __init__(self, body):
        self.body = body

    def __repr__(self):
        return "union{%s}" % self.body



#!/usr/bin/env python3
#
# SPDX-License-Identifier: LGPL-2.1-or-later
#
# A "swiss army knife" tool for qemu capability probing '.replies' files. See
# below in 'description' for more information.

from pathlib import Path
import argparse
import json
import os
import sys
import re


class qrtException(Exception):
    pass


class qmpSchemaException(Exception):
    pass


# Load the 'replies' file into a list of (command, reply) tuples of parsed JSON
def qemu_replies_load(filename):
    conv = []

    with open(filename, "r") as fh:
        command = None
        jsonstr = ''

        try:
            for line in fh:
                jsonstr += line

                if line == '\n':
                    if command is None:
                        command = json.loads(jsonstr)
                    else:
                        conv.append({'cmd': command, 'rep': json.loads(jsonstr)})
                        command = None

                    jsonstr = ''

            if command is not None and jsonstr != '':
                conv.append({'cmd': command, 'rep': json.loads(jsonstr)})
                command = None
                jsonstr = ''

        except json.decoder.JSONDecodeError as je:
            raise qrtException("JSON error:\n'%s'\nwhile processing snippet:\n'%s'" % (je, jsonstr))

        if command is not None or jsonstr != '':
            if command is not None:
                errorstr = json.dumps(command, indent=2)
            else:
                errorstr = jsonstr

            raise qrtException("replies file error: Missing reply for command:\n'%s'" % errorstr)

    return conv


# Format the list of (command, reply) tuples into a string and compare it with
# the 'replies' file. Optionally regenerate the replies file if the output doesn't match
def qemu_replies_compare_or_replace(filename, conv, regenerate_on_error):
    actual = ''
    seq = 9999  # poison the initial counter state

    # possibly fix mis-ordererd 'id' fields
    for c in conv:
        # 'qmp_capabilities' command restarts the numbering sequence
        if c['cmd']['execute'] == 'qmp_capabilities':
            seq = 1

        newid = 'libvirt-%d' % seq
        c['cmd']['id'] = newid
        c['rep']['id'] = newid

        seq += 1

        # format the output string
        if len(actual) != 0:
            actual += '\n\n'

        actual += json.dumps(c['cmd'], indent=2) + '\n\n' + json.dumps(c['rep'], indent=2)

    expect = ''
    actual += '\n'

    with open(filename, "r") as fh:
        expect = fh.read()

    if actual != expect:
        if regenerate_on_error:
            with open(filename, "w") as fh:
                fh.write(actual)

        raise qrtException("replies file error: Expected content of '%s' doesn't match actual content" % filename)


# Process the replies file programmatically here.
# The 'conv' argument contains the whole conversation as a list of
# (command, reply) tuples, where both command and reply are already parsed JSON
# and thus represented by native python types (dict, list, etc ...)
#
# The code below contains a few examples and hints how to use the programatic
# processing. Do not forget to use '--regenerate' flag to update the output files.
#
# Beware that this updates the output file which is used as input for any
# subsequent re-run of the tool which can re-apply the modification.
def modify_replies(conv):
    return  # remove this to enable modifications

    version = None  # filled with a dictionary  with 'major', 'minor', 'micro' keys

    # find version of current qemu for later use
    for c in conv:
        if c['cmd']['execute'] == 'query-version':
            version = c['rep']['return']['qemu']
            break

    if version is None:
        raise Exception("'query-version' not found in the .replies file")

    idx = -1
    # Find index of a command, in this case we're looking for the last
    # invocation of given command
    for i in range(len(conv)):
        c = conv[i]

        if c['cmd']['execute'] == 'device-list-properties':
            idx = i

    if idx == -1:
        raise Exception("entry not found")

    # Prepare data for inserting a new command

    # Command definition and error are instantiated via native python types
    cmd = {'execute': 'device-list-properties',
           'arguments': {'typename': 'example-device'}}

    reply_unsupp = {'error': {'class': 'DeviceNotFound',
                              'desc': "Device 'example-device' not found"}}

    # Real reply data can be also parsed from JSON
    reply = json.loads('''
    {
      "return": [
        {
          "name": "dummy_prop",
          "type": "str"
        },
        {
          "name": "test",
          "type": "str"
        }
        ]
    }
    ''')

    # insert command into the QMP conversation based on version of qemu
    if version['major'] >= 8 and version['minor'] > 0:
        conv.insert(idx, {'cmd': cmd, 'rep': reply})
    else:
        conv.insert(idx, {'cmd': cmd, 'rep': reply_unsupp})


# Validates that 'entry' (an member of the QMP schema):
# - checks that it's a Dict (imported from a JSON object)
# - checks that all 'mandatory' fields are present and their types match
# - checks the types of all 'optional' fields
# - checks that no unknown fields are present
def validate_qmp_schema_check_keys(entry, mandatory, optional):
    keys = set(entry.keys())

    for k, t in mandatory:
        try:
            keys.remove(k)
        except KeyError:
            raise qmpSchemaException("missing mandatory key '%s' in schema '%s'" % (k, entry))

        if not isinstance(entry[k], t):
            raise qmpSchemaException("key '%s' is not of the expected type '%s' in schema '%s'" % (k, t, entry))

    for k, t in optional:
        if k in keys:
            keys.discard(k)

            if not isinstance(entry[k], t):
                raise qmpSchemaException("key '%s' is not of the expected type '%s' in schema '%s'" % (k, t, entry))

    if len(keys) > 0:
        raise qmpSchemaException("unhandled keys '%s' in schema '%s'" % (','.join(list(keys)), entry))


# Validates the optional 'features' and that they consist only of strings
def validate_qmp_schema_check_features_list(entry):
    for f in entry.get('features', []):
        if not isinstance(f, str):
            raise qmpSchemaException("broken 'features' list in schema entry '%s'" % entry)


# Validate that the passed schema has only members supported by this script and
# by the libvirt internals. This is useful to stay up to date with any changes
# to the schema.
def validate_qmp_schema(schemalist):
    for entry in schemalist:
        if not isinstance(entry, dict):
            raise qmpSchemaException("schema entry '%s' is not a JSON Object (dict)" % (entry))

        if entry.get('meta-type', None) == 'command':
            validate_qmp_schema_check_keys(entry,
                                           mandatory=[('name', str),
                                                      ('meta-type', str),
                                                      ('arg-type', str),
                                                      ('ret-type', str)],
                                           optional=[('features', list),
                                                     ('allow-oob', bool)])

            validate_qmp_schema_check_features_list(entry)

        elif entry.get('meta-type', None) == 'event':
            validate_qmp_schema_check_keys(entry,
                                           mandatory=[('name', str),
                                                      ('meta-type', str),
                                                      ('arg-type', str)],
                                           optional=[('features', list)])

            validate_qmp_schema_check_features_list(entry)

        elif entry.get('meta-type', None) == 'object':
            validate_qmp_schema_check_keys(entry,
                                           mandatory=[('name', str),
                                                      ('meta-type', str),
                                                      ('members', list)],
                                           optional=[('tag', str),
                                                     ('variants', list),
                                                     ('features', list)])

            validate_qmp_schema_check_features_list(entry)

            for m in entry.get('members', []):
                validate_qmp_schema_check_keys(m,
                                               mandatory=[('name', str),
                                                          ('type', str)],
                                               optional=[('default', type(None)),
                                                         ('features', list)])
                validate_qmp_schema_check_features_list(m)

            for m in entry.get('variants', []):
                validate_qmp_schema_check_keys(m,
                                               mandatory=[('case', str),
                                                          ('type', str)],
                                               optional=[])

        elif entry.get('meta-type', None) == 'array':
            validate_qmp_schema_check_keys(entry,
                                           mandatory=[('name', str),
                                                      ('meta-type', str),
                                                      ('element-type', str)],
                                           optional=[])

        elif entry.get('meta-type', None) == 'enum':
            validate_qmp_schema_check_keys(entry,
                                           mandatory=[('name', str),
                                                      ('meta-type', str)],
                                           optional=[('members', list),
                                                     ('values', list)])

            for m in entry.get('members', []):
                validate_qmp_schema_check_keys(m,
                                               mandatory=[('name', str)],
                                               optional=[('features', list)])
                validate_qmp_schema_check_features_list(m)

        elif entry.get('meta-type', None) == 'alternate':
            validate_qmp_schema_check_keys(entry,
                                           mandatory=[('name', str),
                                                      ('meta-type', str),
                                                      ('members', list)],
                                           optional=[])

            for m in entry.get('members', []):
                validate_qmp_schema_check_keys(m,
                                               mandatory=[('type', str)],
                                               optional=[])

        elif entry.get('meta-type', None) == 'builtin':
            validate_qmp_schema_check_keys(entry,
                                           mandatory=[('name', str),
                                                      ('meta-type', str),
                                                      ('json-type', str)],
                                           optional=[])

        else:
            raise qmpSchemaException("unknown or missing 'meta-type' in schema entry '%s'" % entry)


# Recursively traverse the schema and print out the schema query strings for
# the corresponding entries. In certain cases the schema references itself,
# which is handled by passing a 'trace' list which contains the current path
def dump_qmp_probe_strings_iter(name, cur, trace, schema):
    obj = schema[name]

    if name in trace:
        # The following is not a query string but sometimes useful for debugging
        # print('%s (recursion)' % cur)
        return

    trace = trace + [name]

    if obj['meta-type'] == 'command' or obj['meta-type'] == 'event':
        arguments = obj.get('arg-type', None)
        returns = obj.get('ret-type', None)

        print(cur)

        for f in obj.get('features', []):
            print('%s/$%s' % (cur, f))

        if arguments:
            dump_qmp_probe_strings_iter(arguments, cur + '/arg-type', trace, schema)

        if returns:
            dump_qmp_probe_strings_iter(returns, cur + '/ret-type', trace, schema)

    elif obj['meta-type'] == 'object':
        members = sorted(obj.get('members', []), key=lambda d: d['name'])
        variants = sorted(obj.get('variants', []), key=lambda d: d['case'])

        for f in obj.get('features', []):
            print('%s/$%s' % (cur, f))

        for memb in members:
            membpath = "%s/%s" % (cur, memb['name'])
            print(membpath)

            # object members can be queried for optionality by '*'
            if 'default' in memb:
                print("%s/*%s" % (cur, memb['name']))

            for f in memb.get('features', []):
                print('%s/$%s' % (membpath, f))

            dump_qmp_probe_strings_iter(memb['type'], membpath, trace, schema)

        for var in variants:
            varpath = "%s/+%s" % (cur, var['case'])
            print(varpath)
            dump_qmp_probe_strings_iter(var['type'], varpath, trace, schema)

    elif obj['meta-type'] == 'enum':
        members = sorted(obj.get('members', []), key=lambda d: d['name'])

        for m in members:
            print('%s/^%s' % (cur, m['name']))

            for f in m.get('features', []):
                print('%s/^%s/$%s' % (cur, m['name'], f))

    elif obj['meta-type'] == 'array':
        dump_qmp_probe_strings_iter(obj['element-type'], cur, trace, schema)

    elif obj['meta-type'] == 'builtin':
        print('%s/!%s' % (cur, name))

    elif obj['meta-type'] == 'alternate':
        for var in obj['members']:
            dump_qmp_probe_strings_iter(var['type'], cur, trace, schema)


def dump_qmp_probe_strings(schemalist, dumpprefix):
    schemadict = {}
    toplevel = []

    for memb in schemalist:
        schemadict[memb['name']] = memb

        if memb['meta-type'] == 'command' or memb['meta-type'] == 'event':
            toplevel.append(memb['name'])

    toplevel.sort()

    for c in toplevel:
        dump_qmp_probe_strings_iter(c, dumpprefix + '(qmp) ' + c, [], schemadict)


def dump_qom_list_types(conv, dumpprefix):
    types = []

    for c in conv:
        if c['cmd']['execute'] == 'qom-list-types':
            for qomtype in c['rep']['return']:
                # validate known fields:
                # 'parent' is ignored below as it causes output churn
                for k in qomtype:
                    if k not in ['name', 'parent']:
                        raise Exception("Unhandled 'qom-list-types' field '%s'" % k)

                types.append(qomtype['name'])

            c['processed'] = True

            break

    types.sort()

    for t in types:
        print(dumpprefix + '(qom) ' + t)


def dump_device_and_object_properties(conv, dumpprefix):
    ent = []

    for c in conv:
        prefix = None

        if c['cmd']['execute'] == 'device-list-properties':
            prefix = '(dev-prop)'

        if c['cmd']['execute'] == 'qom-list-properties':
            prefix = '(qom-prop)'

        if prefix is None:
            continue

        c['processed'] = True

        if 'return' not in c['rep']:
            continue

        for arg in c['rep']['return']:
            for k in arg:
                if k not in ['name', 'type', 'description', 'default-value']:
                    raise Exception("Unhandled 'device-list-properties'/'qom-list-properties' typename '%s' field '%s'" % (c['cmd']['arguments']['typename'], k))

            if 'default-value' in arg:
                defval = ' (%s)' % str(arg['default-value'])
            else:
                defval = ''

            ent.append('%s %s %s %s%s' % (prefix,
                                          c['cmd']['arguments']['typename'],
                                          arg['name'],
                                          arg['type'],
                                          defval))
    ent.sort()

    for e in ent:
        print(dumpprefix + e)


# Sort helper for version string e.g. '11.0', '1.2' etc. Tolerates empty version.
def machine_type_sorter(item):
    key = item[0]

    if key == '':
        return [0]

    return list(map(int, key.split('.')))


def dump_machine_types(conv, dumpprefix):
    machines = dict()
    aliases = []
    dumped_kvm = False

    for c in conv:
        if c['cmd']['execute'] == 'query-machines':

            c['processed'] = True

            if dumped_kvm:
                continue

            for machine in c['rep']['return']:
                deprecated = False
                name = machine['name']
                version = ''
                match = re.fullmatch(r'(.+)-(\d+\.\d+)', name)

                if match is not None:
                    name = match.group(1)
                    version = match.group(2)

                if 'deprecated' in machine:
                    deprecated = machine['deprecated']

                if 'alias' in machine:
                    aliases.append('%s -> %s' % (machine['alias'], machine['name']))

                if name not in machines:
                    machines[name] = {}

                machines[name][version] = deprecated

                # Dump only the machines for the first occurence of 'query-machines'
                dumped_kvm = True

    for (machine, versions) in sorted(machines.items()):
        for (version, deprecated) in sorted(versions.items(), key=machine_type_sorter):
            d = ''
            if deprecated:
                d = ' (deprecated)'

            if len(version) > 0:
                version = '-' + version

            print('(machine) %s%s%s' % (machine, version, d))

    aliases.sort()

    for a in aliases:
        print(dumpprefix + '(machine alias) ' + a)


def dump_command_line_options(c, dumpprefix):
    optpar = []

    for opt in c['rep']['return']:
        for par in opt['parameters']:
            optpar.append('%s %s' % (opt['option'], par['name']))

    optpar.sort()

    for o in optpar:
        print(dumpprefix + '(cl-opt) ' + o)


def dump_other(conv, dumpprefix):
    for c in conv:
        if c['cmd']['execute'] == 'query-version':
            print('%s(version) %s.%s.%s %s' % (dumpprefix,
                                               c['rep']['return']['qemu']['major'],
                                               c['rep']['return']['qemu']['minor'],
                                               c['rep']['return']['qemu']['micro'],
                                               c['rep']['return']['package']))
            c['processed'] = True

        if c['cmd']['execute'] == 'query-target':
            print('%s(target) %s' % (dumpprefix, c['rep']['return']['arch']))
            c['processed'] = True

        if c['cmd']['execute'] == 'query-kvm':
            print('%s(kvm) present:%s enabled:%s' % (dumpprefix,
                                                     c['rep']['return']['present'],
                                                     c['rep']['return']['enabled']))
            c['processed'] = True

        if c['cmd']['execute'] == 'query-command-line-options':
            dump_command_line_options(c, dumpprefix)
            c['processed'] = True


# dumps the parts of the .replies file which are not handled by the various dump_
# helpers
def dump_unprocessed(conv):
    actual = ''

    for c in conv:
        if 'processed' in c and c['processed'] is True:
            continue

        # skip stuf not making sense to be processed:
        # 'qmp_capabilities' - startup of QMP, no interesting data
        # 'query-cpu-model-expansion' - too host dependant, nothing relevant
        if c['cmd']['execute'] in ['qmp_capabilities', 'query-cpu-model-expansion']:
            continue

        # skip commands not having successful return
        if 'return' not in c['rep']:
            continue

        actual += json.dumps(c['cmd'], indent=2) + '\n\n' + json.dumps(c['rep'], indent=2)

    if actual != '':
        for line in actual.split('\n'):
            print('(unprocessed) ' + line)


def process_one(filename, args):
    try:
        conv = qemu_replies_load(filename)
        dumped = False
        dumpprefix = ''

        if args.repliesdir:
            dumpprefix = filename + ': '

        modify_replies(conv)

        for c in conv:
            if c['cmd']['execute'] == 'query-qmp-schema':
                validate_qmp_schema(c['rep']['return'])

                if args.dump_all or args.dump_qmp_query_strings:
                    dump_qmp_probe_strings(c['rep']['return'], dumpprefix)
                    c['processed'] = True
                    dumped = True

        if args.dump_all:
            dump_other(conv, dumpprefix)
            dump_qom_list_types(conv, dumpprefix)
            dump_device_and_object_properties(conv, dumpprefix)
            dump_machine_types(conv, dumpprefix)
            dumped = True

        if args.dump_unprocessed:
            dump_unprocessed(conv)
            dumped = True

        if dumped:
            return True

        qemu_replies_compare_or_replace(filename, conv, args.regenerate)

    except qrtException as e:
        print("'%s' ... FAIL\n%s" % (filename, e))
        return False
    except qmpSchemaException as qe:
        print("'%s' ... FAIL\nqmp schema error: %s" % (filename, qe))
        return False

    print("'%s' ... OK" % filename)
    return True


description = '''A Swiss army knife tool for '.replies' files used by 'qemucapabilitiestest'

This tool is used to validate, programmatically update or inspect the
'.*replies' normally stored files under 'tests/qemucapabilitiesdata'.

By default the file(s) passed as positional argument are used. All '.replies'
files in a directory can be processed by specifying '--repliesdir /path/to/dir'
argument.

The default mode is validation which checks the following:
    - each command has a reply and both are valid JSON
    - numbering of the 'id' field is as expected
    - the input file has the expected JSON formatting
    - the QMP schema from qemu is fully covered by libvirt's code

In 'dump' mode if '-dump-all' or one of the specific '-dump-*' flags (below)
is selected the script outputs information gathered from the given '.replies'
file. The data is also usable for comparing two '.replies' files in a "diffable"
fashion as many of the query commands may change ordering or naming without
functional impact on libvirt. The following specific dump options are useful
on it's own:

  --dump-qmp-query-strings

    Dumps all possible valid QMP capability query strings based on the current
    qemu version in format used by virQEMUQAPISchemaPathGet or
    virQEMUCapsQMPSchemaQueries. It's useful to find specific query string
    without having to piece the information together from 'query-qmp-schema'

The tool can be also used to programmaticaly modify the '.replies' file by
editing the 'modify_replies' method directly in the source, or for
re-formatting and re-numbering the '.replies' file to conform with the required
format. To update the output file the '--regenerate' flag can be used or the
'VIR_TEST_REGENERATE_OUTPUT' environment variable must be set to '1'.
'''

if os.environ.get('VIR_TEST_REGENERATE_OUTPUT', '0') == '1':
    default_regenerate = True
else:
    default_regenerate = False

parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                 description=description)

parser.add_argument('--regenerate', action="store_true", default=default_regenerate,
                    help="regenerate output file if actual output doesn't match")

parser.add_argument('--repliesdir', default='',
                    help='use all .replies files from the directory')

parser.add_argument('replyfiles', nargs='*',
                    help='.replies file(s) to process')

parser.add_argument('--dump-all', action='store_true',
                    help='invoke all --dump-* sub-commands')

parser.add_argument('--dump-qmp-query-strings', action='store_true',
                    help='dump QMP schema in form of query strings used to probe capabilities')


parser.add_argument('--dump-unprocessed', action='store_true',
                    help='dump JSON of commands unprocessed by any of the --dump-* options')

args = parser.parse_args()

files = []

if args.replyfiles:
    files += args.replyfiles

if args.repliesdir:
    files += Path(args.repliesdir).glob('*.replies')

if len(files) == 0:
    parser.print_help()
    sys.exit(1)

fail = False

for file in files:
    if not process_one(str(file), args):
        fail = True

if fail:
    sys.exit(1)



#!/usr/bin/env python3

import re
import sys

if len(sys.argv) != 2:
    print(f"usage: {sys.argv[0]} POTFILE", file=sys.stderr)
    sys.exit(1)

potfile = sys.argv[1]

failed = 0


def print_msg(files, msgs):
    if len(msgs) == 0:
        return

    print("\n".join(files))

    for m in msgs:
        print(f"  {m}")

    global failed
    failed += 1


with open(potfile, "r") as pot:
    files = []
    msgs = []
    cFormat = False

    for line in pot:
        if not line or line.startswith("msgstr "):
            print_msg(files, msgs)
            files = []
            msgs = []
            cFormat = False
            continue

        if line.startswith("#: "):
            files.extend(line[3:].split())
            continue

        if line.startswith("#,"):
            cFormat = " c-format" in line
            continue

        m = re.search(r'^(msgid )?"(.*%[^%$ ]*[a-zA-Z].*)"', line)
        if cFormat and m is not None:
            msgs.append(m.group(2))

if failed:
    print(f"Found {failed} messages without permutable format strings!",
          file=sys.stderr)
    sys.exit(1)
